==> trimica-1.3.2%2B1.21.8-fabric-sources.jar: com/bawnorton/trimica/client/texture/AbstractTrimSpriteFactory.java <==
package com.bawnorton.trimica.client.texture;

import com.bawnorton.trimica.client.palette.AnimatedTrimPalette;
import com.bawnorton.trimica.client.palette.TrimPalette;
import com.bawnorton.trimica.platform.Platform;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import it.unimi.dsi.fastutil.ints.IntList;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.nio.file.Path;
import java.util.*;
import net.minecraft.class_1011;
import net.minecraft.class_10539;
import net.minecraft.class_1079;
import net.minecraft.class_2960;
import net.minecraft.class_7368;
import net.minecraft.class_7677;
import net.minecraft.class_7764;
import net.minecraft.class_7771;
import net.minecraft.class_8053;
import net.minecraft.class_9473;

public abstract class AbstractTrimSpriteFactory {
	static final Map<class_2960, class_10539> textureCache = new HashMap<>();

	protected final int width;
	protected final int height;

	protected AbstractTrimSpriteFactory(int width, int height) {
		this.width = width;
		this.height = height;
	}

	//? if >=1.21.10 {
	/*public TrimSpriteContents create(ResourceLocation texture, @Nullable ArmorTrim trim, @Nullable DataComponentGetter componentGetter) {
		if (trim == null) {
			return TrimSpriteContents.noPalette(new SpriteContents(texture, new FrameSize(width, height), empty()));
		}
		TrimSpriteMetadata metadata = getSpriteMetadata(trim, componentGetter, texture);
		if (metadata == null) {
			return TrimSpriteContents.noPalette(new SpriteContents(texture, new FrameSize(width, height), empty()));
		}
		NativeImage image = createImageFromMetadata(metadata);
		ResourceMetadata resourceMetadata = getResourceMetadataFromSpriteMetadata(metadata);
		return new TrimSpriteContents(
				new SpriteContents(
						texture,
						new FrameSize(width, height),
						image,
						resourceMetadata.getSection(AnimationMetadataSection.TYPE),
						resourceMetadata.getTypedSections(List.of(AnimationMetadataSection.TYPE))
				),
				metadata.palette()
		);
	}
	*///?} else {
	public TrimSpriteContents create(class_2960 texture, @Nullable class_8053 trim, @Nullable class_9473 componentGetter) {
		if (trim == null) {
			return TrimSpriteContents.noPalette(new class_7764(texture, new class_7771(width, height), empty(), class_7368.field_38688));
		}
		TrimSpriteMetadata metadata = getSpriteMetadata(trim, componentGetter, texture);
		if (metadata == null) {
			return TrimSpriteContents.noPalette(new class_7764(texture, new class_7771(width, height), empty(), class_7368.field_38688));
		}
		class_1011 image = createImageFromMetadata(metadata);
		class_7368 resourceMetadata = getResourceMetadataFromSpriteMetadata(metadata);
		return new TrimSpriteContents(new class_7764(texture, new class_7771(width, height), image, resourceMetadata), metadata.palette());
	}
	//?}

	@Nullable
	protected abstract TrimSpriteMetadata getSpriteMetadata(class_8053 trim, @Nullable class_9473 componentGetter, class_2960 texture);

	protected abstract class_1011 createImageFromMetadata(TrimSpriteMetadata metadata);

	private @NotNull class_7368 getResourceMetadataFromSpriteMetadata(TrimSpriteMetadata metadata) {
		if (metadata.isAnimated()) {
			return new class_7368() {
				@SuppressWarnings("unchecked")
				@Override
				public <T> @NotNull Optional<T> method_43041(@NotNull class_7677<T> metadataSectionType) {
					if (metadataSectionType.equals(class_1079.field_55537)) {
						return Optional.of((T) new class_1079(
								Optional.empty(),
								Optional.empty(),
								Optional.empty(),
								1,
								false
						));
					}
					return Optional.empty();
				}
			};
		} else {
			return class_7368.field_38688;
		}
	}

	protected class_1011 createColouredImage(TrimSpriteMetadata metadata, class_10539 contents) {
		TrimPalette palette = metadata.palette();
		class_1011 grey = contents.comp_3447();
		class_1011 coloured;
		Path debug = Platform.getDebugDirectory().resolve(metadata.toDebugPath());
		if (Platform.isDev()) {
			try {
				debug.toFile().mkdirs();
				grey.method_4314(debug.resolve("greyscale.png"));
			} catch (IOException e) {
				throw new RuntimeException("Failed to write greyscale pattern image", e);
			}
		}
		if (palette.isAnimated()) {
			coloured = createColouredPatternAnimation(grey, palette);
		} else {
			coloured = createColouredPatternImage(grey, palette.getColours(), palette.isBuiltin());
		}
		if (Platform.isDev()) {
			try {
				coloured.method_4314(debug.resolve("coloured.png"));
			} catch (IOException e) {
				throw new RuntimeException("Failed to write coloured pattern image", e);
			}
		}
		return coloured;
	}

	protected class_1011 createColouredPatternImage(class_1011 greyscalePatternImage, List<Integer> colours, boolean builtin) {
		class_1011 coloured = new class_1011(width, height, false);
		IntList[] colourPositions = new IntList[256];

		int count = 0;
		boolean encounteredTransparent = false;
		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				int pixel = greyscalePatternImage.method_61940(x, y);
				int alpha = (pixel >> 24) & 0xFF;
				if (alpha == 0) {
					coloured.method_61941(x, y, 0);
					encounteredTransparent = true;
					continue;
				}

				int shade = pixel & 0xFF;
				IntList positions = colourPositions[shade];
				if (positions == null) {
					positions = new IntArrayList();
					colourPositions[shade] = positions;
					count++;
				}
				positions.add(x);
				positions.add(y);
			}
		}
		if (count == 0) {
			return empty();
		}
		if (encounteredTransparent) {
			count++;
		}

		int paletteIndex = Math.min(count - 1, TrimPalette.PALETTE_SIZE - 1);
		for (int colour = 0; colour < 256; colour++) {
			IntList positions = colourPositions[colour];
			if (positions == null) continue;

			int paletteColour;

			if (colour == 0) {
				paletteColour = 0;
			} else {
				if (paletteIndex > 0) {
					paletteColour = colours.get(paletteIndex) | (0xFF << 24);
				} else {
					paletteColour = 0;
				}
				paletteIndex--;
			}

			for (int i = 0; i < positions.size(); i += 2) {
				int x = positions.getInt(i);
				int y = positions.getInt(i + 1);
				coloured.method_61941(x, y, builtin ? paletteColour : applyGrayscaleMask(paletteColour, colour));
			}
		}
		return coloured;
	}

	protected class_1011 createColouredPatternAnimation(class_1011 image, TrimPalette palette) {
		List<List<Integer>> frames = getFrames(palette);
		int frameHeight = frames.size();
		int frameWidth = image.method_4307();
		int totalHeight = frameHeight * image.method_4323();
		class_1011 stitchedImage = new class_1011(frameWidth, totalHeight, false);
		for (int frameCount = 0; frameCount < frames.size(); frameCount++) {
			List<Integer> frame = frames.get(frameCount);
			class_1011 coloured = createColouredPatternImage(image, frame, palette.isBuiltin());
			for (int y = 0; y < image.method_4323(); y++) {
				for (int x = 0; x < image.method_4307(); x++) {
					int pixel = coloured.method_61940(x, y);
					stitchedImage.method_61941(x, y + (frameCount * image.method_4323()), pixel);
				}
			}
			coloured.close();
		}
		return stitchedImage;
	}

	private @NotNull List<List<Integer>> getFrames(TrimPalette palette) {
		List<Integer> interpolatedColours = palette instanceof AnimatedTrimPalette animated ? animated.getAnimationColours() : palette.getColours();
		List<List<Integer>> frames = new ArrayList<>();
		for (int i = 0; i < AnimatedTrimPalette.ANIMATED_PALETTE_SIZE; i++) {
			List<Integer> frame = new ArrayList<>();
			for (int j = 0; j < TrimPalette.PALETTE_SIZE; j++) {
				int index = (i + j) % interpolatedColours.size();
				int colour = interpolatedColours.get(index);
				frame.add(colour);
			}
			frames.add(frame);
		}
		return frames;
	}

	private int applyGrayscaleMask(int colorABGR, int maskABGR) {
		int brightness = maskABGR & 0xFF;

		int alpha = (colorABGR >> 24) & 0xFF;
		int blue = (colorABGR >> 16) & 0xFF;
		int green = (colorABGR >> 8) & 0xFF;
		int red = colorABGR & 0xFF;

		if (brightness < 128) {
			red = (red * brightness + 64) >> 7;
			green = (green * brightness + 64) >> 7;
			blue = (blue * brightness + 64) >> 7;
		}

		return (alpha << 24) | (blue << 16) | (green << 8) | red;
	}

	public void clear() {
		textureCache.forEach((k, v) -> v.close());
		textureCache.clear();
	}

	protected class_1011 empty() {
		class_1011 image = new class_1011(width, height, false);
		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				image.method_61941(x, y, 0);
			}
		}
		return image;
	}
}

==> trimica-1.4.3%2B1.21.10-fabric-sources.jar: com/bawnorton/trimica/client/texture/AbstractTrimSpriteFactory.java <==
package com.bawnorton.trimica.client.texture;

import com.bawnorton.trimica.client.palette.AnimatedTrimPalette;
import com.bawnorton.trimica.client.palette.TrimPalette;
import com.bawnorton.trimica.platform.Platform;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import it.unimi.dsi.fastutil.ints.IntList;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.nio.file.Path;
import java.util.*;
import net.minecraft.class_1011;
import net.minecraft.class_10539;
import net.minecraft.class_1079;
import net.minecraft.class_2960;
import net.minecraft.class_7368;
import net.minecraft.class_7677;
import net.minecraft.class_7764;
import net.minecraft.class_7771;
import net.minecraft.class_8053;
import net.minecraft.class_9473;

public abstract class AbstractTrimSpriteFactory {
	static final Map<class_2960, class_10539> textureCache = new HashMap<>();

	protected final int width;
	protected final int height;

	protected AbstractTrimSpriteFactory(int width, int height) {
		this.width = width;
		this.height = height;
	}

	//? if >=1.21.10 {
	public TrimSpriteContents create(class_2960 texture, @Nullable class_8053 trim, @Nullable class_9473 componentGetter) {
		if (trim == null) {
			return TrimSpriteContents.noPalette(new class_7764(texture, new class_7771(width, height), empty()));
		}
		TrimSpriteMetadata metadata = getSpriteMetadata(trim, componentGetter, texture);
		if (metadata == null) {
			return TrimSpriteContents.noPalette(new class_7764(texture, new class_7771(width, height), empty()));
		}
		class_1011 image = createImageFromMetadata(metadata);
		class_7368 resourceMetadata = getResourceMetadataFromSpriteMetadata(metadata);
		return new TrimSpriteContents(
				new class_7764(
						texture,
						new class_7771(width, height),
						image,
						resourceMetadata.method_43041(class_1079.field_55537),
						resourceMetadata.method_72364(List.of(class_1079.field_55537))
				),
				metadata.palette()
		);
	}
	//?} else {
	/*public TrimSpriteContents create(ResourceLocation texture, @Nullable ArmorTrim trim, @Nullable DataComponentGetter componentGetter) {
		if (trim == null) {
			return TrimSpriteContents.noPalette(new SpriteContents(texture, new FrameSize(width, height), empty(), ResourceMetadata.EMPTY));
		}
		TrimSpriteMetadata metadata = getSpriteMetadata(trim, componentGetter, texture);
		if (metadata == null) {
			return TrimSpriteContents.noPalette(new SpriteContents(texture, new FrameSize(width, height), empty(), ResourceMetadata.EMPTY));
		}
		NativeImage image = createImageFromMetadata(metadata);
		ResourceMetadata resourceMetadata = getResourceMetadataFromSpriteMetadata(metadata);
		return new TrimSpriteContents(new SpriteContents(texture, new FrameSize(width, height), image, resourceMetadata), metadata.palette());
	}
	*///?}

	@Nullable
	protected abstract TrimSpriteMetadata getSpriteMetadata(class_8053 trim, @Nullable class_9473 componentGetter, class_2960 texture);

	protected abstract class_1011 createImageFromMetadata(TrimSpriteMetadata metadata);

	private @NotNull class_7368 getResourceMetadataFromSpriteMetadata(TrimSpriteMetadata metadata) {
		if (metadata.isAnimated()) {
			return new class_7368() {
				@SuppressWarnings("unchecked")
				@Override
				public <T> @NotNull Optional<T> method_43041(@NotNull class_7677<T> metadataSectionType) {
					if (metadataSectionType.equals(class_1079.field_55537)) {
						return Optional.of((T) new class_1079(
								Optional.empty(),
								Optional.empty(),
								Optional.empty(),
								1,
								false
						));
					}
					return Optional.empty();
				}
			};
		} else {
			return class_7368.field_38688;
		}
	}

	protected class_1011 createColouredImage(TrimSpriteMetadata metadata, class_10539 contents) {
		TrimPalette palette = metadata.palette();
		class_1011 grey = contents.comp_3447();
		class_1011 coloured;
		Path debug = Platform.getDebugDirectory().resolve(metadata.toDebugPath());
		if (Platform.isDev()) {
			try {
				debug.toFile().mkdirs();
				grey.method_4314(debug.resolve("greyscale.png"));
			} catch (IOException e) {
				throw new RuntimeException("Failed to write greyscale pattern image", e);
			}
		}
		if (palette.isAnimated()) {
			coloured = createColouredPatternAnimation(grey, palette);
		} else {
			coloured = createColouredPatternImage(grey, palette.getColours(), palette.isBuiltin());
		}
		if (Platform.isDev()) {
			try {
				coloured.method_4314(debug.resolve("coloured.png"));
			} catch (IOException e) {
				throw new RuntimeException("Failed to write coloured pattern image", e);
			}
		}
		return coloured;
	}

	protected class_1011 createColouredPatternImage(class_1011 greyscalePatternImage, List<Integer> colours, boolean builtin) {
		class_1011 coloured = new class_1011(width, height, false);
		IntList[] colourPositions = new IntList[256];

		int count = 0;
		boolean encounteredTransparent = false;
		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				int pixel = greyscalePatternImage.method_61940(x, y);
				int alpha = (pixel >> 24) & 0xFF;
				if (alpha == 0) {
					coloured.method_61941(x, y, 0);
					encounteredTransparent = true;
					continue;
				}

				int shade = pixel & 0xFF;
				IntList positions = colourPositions[shade];
				if (positions == null) {
					positions = new IntArrayList();
					colourPositions[shade] = positions;
					count++;
				}
				positions.add(x);
				positions.add(y);
			}
		}
		if (count == 0) {
			return empty();
		}
		if (encounteredTransparent) {
			count++;
		}

		int paletteIndex = Math.min(count - 1, TrimPalette.PALETTE_SIZE - 1);
		for (int colour = 0; colour < 256; colour++) {
			IntList positions = colourPositions[colour];
			if (positions == null) continue;

			int paletteColour;

			if (colour == 0) {
				paletteColour = 0;
			} else {
				if (paletteIndex > 0) {
					paletteColour = colours.get(paletteIndex) | (0xFF << 24);
				} else {
					paletteColour = 0;
				}
				paletteIndex--;
			}

			for (int i = 0; i < positions.size(); i += 2) {
				int x = positions.getInt(i);
				int y = positions.getInt(i + 1);
				coloured.method_61941(x, y, builtin ? paletteColour : applyGrayscaleMask(paletteColour, colour));
			}
		}
		return coloured;
	}

	protected class_1011 createColouredPatternAnimation(class_1011 image, TrimPalette palette) {
		List<List<Integer>> frames = getFrames(palette);
		int frameHeight = frames.size();
		int frameWidth = image.method_4307();
		int totalHeight = frameHeight * image.method_4323();
		class_1011 stitchedImage = new class_1011(frameWidth, totalHeight, false);
		for (int frameCount = 0; frameCount < frames.size(); frameCount++) {
			List<Integer> frame = frames.get(frameCount);
			class_1011 coloured = createColouredPatternImage(image, frame, palette.isBuiltin());
			for (int y = 0; y < image.method_4323(); y++) {
				for (int x = 0; x < image.method_4307(); x++) {
					int pixel = coloured.method_61940(x, y);
					stitchedImage.method_61941(x, y + (frameCount * image.method_4323()), pixel);
				}
			}
			coloured.close();
		}
		return stitchedImage;
	}

	private @NotNull List<List<Integer>> getFrames(TrimPalette palette) {
		List<Integer> interpolatedColours = palette instanceof AnimatedTrimPalette animated ? animated.getAnimationColours() : palette.getColours();
		List<List<Integer>> frames = new ArrayList<>();
		for (int i = 0; i < AnimatedTrimPalette.ANIMATED_PALETTE_SIZE; i++) {
			List<Integer> frame = new ArrayList<>();
			for (int j = 0; j < TrimPalette.PALETTE_SIZE; j++) {
				int index = (i + j) % interpolatedColours.size();
				int colour = interpolatedColours.get(index);
				frame.add(colour);
			}
			frames.add(frame);
		}
		return frames;
	}

	private int applyGrayscaleMask(int colorABGR, int maskABGR) {
		int brightness = maskABGR & 0xFF;

		int alpha = (colorABGR >> 24) & 0xFF;
		int blue = (colorABGR >> 16) & 0xFF;
		int green = (colorABGR >> 8) & 0xFF;
		int red = colorABGR & 0xFF;

		if (brightness < 128) {
			red = (red * brightness + 64) >> 7;
			green = (green * brightness + 64) >> 7;
			blue = (blue * brightness + 64) >> 7;
		}

		return (alpha << 24) | (blue << 16) | (green << 8) | red;
	}

	public void clear() {
		textureCache.forEach((k, v) -> v.close());
		textureCache.clear();
	}

	protected class_1011 empty() {
		class_1011 image = new class_1011(width, height, false);
		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				image.method_61941(x, y, 0);
			}
		}
		return image;
	}
}
